"""Main game client module. Handles both single player and multiplayer.
Uses the networking module for both (in single player, server is on
the localhost)"""

import socket
import time
import os

import pygame

from pygame.locals import *

from locals import *

import data

from util import Util, render_text, render_image, str_editor, \
clear_text_cache, fade_to_black, set_fade_state, add_cursor
import network_util
from variables import Variables
from log import log_message, error_message

from sound import play_sound

from unit import Unit
from faction import Faction
from level import Level
from turn import Turn
from action import Action

class Client:

    def __init__(self, screen, host, port = PORT_DEFAULT, server_key = None, client_nick = 'Anonymous'):
        self.screen = screen
        self.host = host
        self.port = port
        self.server_key = server_key
        self.client_id = -1
        self.client_nick = client_nick
        return

    def run(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect((self.host,self.port))
        except socket.error, (value,message):
            error_message('Could not connect to server')
            return False

        network_util.ping(sock)

        while (self.client_id == -1):
            self.client_id = network_util.get_client_id(sock, self.client_nick)

        network_util.set_nick(sock, self.client_id, \
        Variables.vdict["player_name"])

        time.sleep(0.1)

        self.chat_log = []
        self.outgoing_chat_msg = ''
        self.chat_on = False
        self.chat_cursor = 0

        self.level = None
        self.faction = None

        self.offset_x = 40
        self.offset_y = 70

        self.current_turn = None

        self.terminated = False

        self.to_be_parsed = []
        self.turn_parsed = True

        self.drag = False
        self.dragpos = (0,0)

        fade_finished = False
        set_fade_state(FADE_STATE_NONE)
        self.fade_target = FADE_STATE_NONE

        clock = pygame.time.Clock()

        while not (self.terminated and fade_finished):

            server_has_events = True
            while (server_has_events):
                event_amount, event = network_util.query_events(sock, \
                self.client_id)
                if (event != None):
                    #log_message('Client processing event: ' + str(event))
                    event_type, empty_string, value = event.partition(' ')
                    if event_type == COMMAND_CHAT:
                        self.chat_log.append(value[0:CHAT_MSG_MAX_LEN])
                    elif event_type == GAME_COMMAND_SET_RES:
                        res_strs = value.split()
                        res = (int(res_strs[0]), int(res_strs[1]), \
                        int(res_strs[2]))
                        self.faction.set_resources(res)
                        self.turn_parsed = True
                    elif event_type == COMMAND_KILLYOURSELF:
                        log_message('Server is closing - disconnecting client')
                        self.terminated = True
                        self.fade_target = FADE_STATE_HALF
                    elif event_type.startswith(ACTION_PREFIX):
                        if self.current_turn == None:
                            #log_message('Client processing action: ' + str(event))
                            if len(self.to_be_parsed) == 0:
                                self.parse_served_game_input(event)
                            else:
                               queued_event = self.to_be_parsed[0]
                               self.parse_served_game_input(queued_event)
                               self.to_be_parsed.remove(queued_event)
                        else:
                            self.to_be_parsed.append(event)
                if event_amount == 0:
                    server_has_events = False

            event_queue = []

            for event in pygame.event.get():
                if event.type == QUIT or \
                (event.type == KEYDOWN and event.key == K_ESCAPE):
                    self.terminate(sock)

                elif self.chat_on:
                    #Direct input to chat
                    if event.type == KEYDOWN:
                        if event.key == K_RETURN:
                            self.process_chat_input(sock, self.outgoing_chat_msg)
                            self.outgoing_chat_msg = ''
                            self.chat_on = False
                            clear_text_cache()
                        else:
                            self.outgoing_chat_msg, self.chat_cursor = \
                            str_editor(self.outgoing_chat_msg, event)
                else:
                    #Direct input outside chat
                    if event.type == KEYDOWN and event.key == K_RETURN:
                        self.chat_on = True
                    elif event.type == MOUSEBUTTONDOWN or \
                    event.type == MOUSEBUTTONUP or event.type == KEYDOWN:
                        #To be processed later if the player can
                        #affect gameplay at the time
                        event_queue.append(event)

            render_image(self.screen, 'game-background.jpg')

            #Getting the level and faction name from server
            #And rendering related data

            if self.level == None:
                level_name, correct_hash = self.ask_for_level(sock).split()
                self.level = Level(level_name)
                if self.level.file_hash == correct_hash:
                    log_message("Level hash matches")
                else:
                    error_message("Level hash doesn't match! " \
                    + "Check your game version.")
                    self.terminate(sock)
            else:
                self.level.render(self.screen, self.offset_x, self.offset_y)
                if self.faction == None:
                    f_color = self.ask_for_faction(sock)
                    if f_color != False:
                        if f_color == NO_AVAILABLE_FACTIONS:
                            self.faction = Faction('observer')
                        else:
                            self.faction = self.level.get_faction(f_color)
                else:
                    if self.faction.get_color() != 'observer':
                        #Player has a non-observer faction
                        if self.current_turn == None:
                            if self.turn_parsed:
                                #Either the game is starting, or a round has ended.
                                self.current_turn = Turn()
                                winner = self.level.new_turn()
                                self.turn_parsed = False
                                self.level.deselect()
                                
                                if winner != '':
                                    self.chat_log.append('Faction ' \
                                    + winner + ' won the game! Congrats!')
                        else:
                            if self.current_turn.active:
                                self.parse_game_input(event_queue)
                            else:
                                #Turn ended - inform server:
                                if self.send_turn(sock):
                                    self.level.reset_last_turn()
                                    self.level.deselect()
                                    #Turn sent succesfully
                                    #The game starts waiting for server events
                                    #telling what really happened
                                    self.current_turn = None
                                else:
                                    error_message('Error sending turn')
                                    self.current_turn = None
                                    self.terminate(sock)

            self.render_gui()

            fade_finished = fade_to_black(self.screen, self.fade_target)

            pygame.display.flip()

            clock.tick(FPS)

        network_util.disconnect(sock, self.client_id)

        sock.close()

        log_message('Client terminated.')

        return True

    def parse_game_input(self, event_queue):
        """Parses game-related input the player gives to the GUI.
        event_queue is a list of pygame events."""

        if self.drag:
            mouse_pos = pygame.mouse.get_pos()
            self.offset_x = self.dragstart_offset_x + mouse_pos[0] \
            - self.dragpos[0]
            self.offset_y = self.dragstart_offset_y + mouse_pos[1] \
            - self.dragpos[1]


        action_queue = []
        for event in event_queue:
            if event.type == MOUSEBUTTONDOWN and event.button == 1:

                #Check GUI buttons
                end_turn_rect = pygame.Rect(660, 8, 108, 26)
                twist_rect = pygame.Rect(500, 8, 108, 26)

                infobox_action = self.level.use_infobox_gui(event.pos, \
                self.faction.get_resources())

                if end_turn_rect.collidepoint(event.pos):
                    action_queue.append(ACTION_END_TURN)

                elif twist_rect.collidepoint(event.pos):
                    action_queue.append(ACTION_TWIST)

                elif infobox_action != None:
                    if infobox_action != True:
                        action_queue.append(infobox_action)

                #Check object selection
                else:
                    self.level.select(event.pos, self.faction.get_color())
            elif event.type == MOUSEBUTTONDOWN and event.button == 3:
                self.drag = True
                self.dragpos = event.pos
                self.dragstart_offset_x = self.offset_x
                self.dragstart_offset_y = self.offset_y
            elif event.type == MOUSEBUTTONUP and event.button == 3:
                self.drag = False

            elif event.type == KEYDOWN and event.key == K_e:
                action_queue.append(ACTION_END_TURN)
            elif event.type == KEYDOWN and event.key == K_t:
                action_queue.append(ACTION_TWIST)
            elif event.type == KEYDOWN and event.key == K_RIGHT:
                action_queue.append(ACTION_MOVE_RIGHT)
            elif event.type == KEYDOWN and event.key == K_LEFT:
                action_queue.append(ACTION_MOVE_LEFT)
            elif event.type == KEYDOWN and event.key == K_UP:
                action_queue.append(ACTION_MOVE_UP)
            elif event.type == KEYDOWN and event.key == K_DOWN:
                action_queue.append(ACTION_MOVE_DOWN)

        for action in action_queue:
            if type(action) != type(ACTION_MOVE_PREFIX):
                #Action is an action object (should be, at least...)
                action = self.level.buy(action, self.faction.get_color())
                if action != None:
                    self.current_turn.add_action(action)
            elif action.startswith(ACTION_MOVE_PREFIX):
                if self.current_turn != None and self.level.selected_object != None:
                    action = self.level.move(action)
                    if action != None:
                        self.current_turn.add_action(action)
            elif action == ACTION_END_TURN:
                self.current_turn.end()
                log_message('Turn ended')
            elif action == ACTION_TWIST:
                action = self.level.toggle_twist(self.faction.get_color())
                if action != None:
                    self.current_turn.add_action(action)
        return

    def parse_served_game_input(self, action):
        log_message('Parsing served input: ' + action)
        values = action.split()
        if action.startswith(ACTION_MOVE_PREFIX):
            self.level.select_by_tile_coords((int(values[1]), int(values[2])))
            self.level.move(values[0])
        elif action.startswith(ACTION_BUY_PREFIX):
            action = Action(values[0], (int(values[1]), int(values[2])), int(values[3]))
            faction = values[4]
            self.level.buy(action, faction)
        elif action.startswith(ACTION_TWIST):
            log_message('Server told me to twist it')
            self.level.toggle_twist()
        return

    def render_gui(self):
        render_image(self.screen, 'game-decoration.png')

        player_name_img = None

        if self.faction != None:
            #Render faction related GUI - resources etc.
            render_image(self.screen, 'res_icon_rock.png', (20, 574))
            render_image(self.screen, 'res_icon_paper.png', (100, 574))
            render_image(self.screen, 'res_icon_scissors.png', (180, 574))
            res_white, res_black, res_red = self.faction.get_resources()
            res_white_img = render_text(str(res_white), 12)
            res_black_img = render_text(str(res_black), 12)
            res_red_img = render_text(str(res_red), 12)
            self.screen.blit(res_white_img, (36, 570))
            self.screen.blit(res_black_img, (116, 570))
            self.screen.blit(res_red_img, (196, 570))


            render_image(self.screen, \
            'emblem_large_' + self.faction.get_color() + '.png', (378, 569))

            render_image(self.screen, \
            'gui_button_end_turn.png', (660, 8))

            if self.current_turn == None or not self.current_turn.get_twist():
                render_image(self.screen, \
                'gui_button_twist_inactive.png', (500, 8))
            else:
                render_image(self.screen, \
                'gui_button_twist_active.png', (500, 8))

            player_name_img = render_text(self.client_nick, 12, \
            self.faction.get_color_rgb())

        else:
            player_name_img = render_text(self.client_nick, 12)

        self.screen.blit(player_name_img, (400, 570))

        if (len(self.chat_log) > 0):
            #Render chat log
            msgs = 0
            index = len(self.chat_log) - msgs - 1
            while msgs < 4 and index >= 0:
               chat_msg = render_text(self.chat_log[index], 12)
               self.screen.blit(chat_msg, (20, 520 - msgs*20))
               msgs += 1
               index = len(self.chat_log) - msgs - 1

        if (self.chat_on):
            #Render chat input field
            chat_msg_display = add_cursor(self.outgoing_chat_msg, self.chat_cursor)
            chat_input = render_text('> ' + chat_msg_display, 12)
            self.screen.blit(chat_input, (20, 540))
        return

    def ask_for_level(self, sock):
        return network_util.network_command(sock, self.client_id, \
        GAME_COMMAND_ASK_LEVEL, '')

    def ask_for_faction(self, sock):
        return network_util.network_command(sock, self.client_id, \
        GAME_COMMAND_ASK_FACTION, '')

    def send_turn(self, sock):
        network_msgs = self.current_turn.get_network_msgs()
        for network_msg in network_msgs:
            success = network_util.network_command(sock, self.client_id, \
            GAME_COMMAND_SEND_TURN, network_msg)
            if success == False:
                return False
        return True

    def terminate(self, sock):
        if self.server_key != None:
            network_util.terminate_server(sock, self.server_key)
        else:
            self.terminated = True
            self.fade_target = FADE_STATE_HALF
        return

    def process_chat_input(self, sock, message):
        message = message.strip()
        if message != '':
            if message.startswith('/nick'):
                self.client_nick = message[6:len(message)]
                self.client_nick = self.client_nick[0:NICK_MAX_LEN].strip()
                network_util.set_nick(sock, self.client_id, \
                self.client_nick)
            else:
                network_util.submit_chat(sock, self.client_id, message)
