"""Game level module"""

import os
import hashlib
import codecs

import pygame

from locals import *

import data

from log import log_message, error_message

from util import is_config_line, render_image

from shrine import Shrine
from base import Base
from faction import Faction
from action import Action
from unit import Unit

class Level:

    def __init__(self, name):
        name = name.strip()
        self.name = name

        self.file_path = data.filepath(os.path.join('levels', \
        name + FILE_EXT_LEVEL))
        level_file = codecs.open(self.file_path, "r", "utf_8")

        m = hashlib.md5()
        m.update(level_file.read())

        self.file_hash = str(m.hexdigest())

        level_file.close()

        log_message('Level hash: ' + self.file_hash)

        self.width = 10
        self.height = 10

        read_data = False
        y = 0

        self.tiles = {}
        self.objects = []
        self.factions = []

        self.selected_object = None
        
        self.twisted = False        

        level_file = codecs.open(self.file_path, "r", "utf_8")

        for line in level_file.xreadlines():
            values = line.split()
            if not is_config_line(line):
                pass
            elif read_data:
                x = 0
                for c in line:
                    if (c == 'W'):
                        self.tiles[(x, y)] = TILE_WALL
                    x += 1
                y += 1
                if (y == self.height):
                    read_data = False
                    log_message('Read level data')
            elif values[0] == 'data':
                read_data = True
            elif values[0] == 'width':
                self.width = int(values[1])
            elif values[0] == 'height':
                self.height = int(values[1])
            elif values[0] == 'faction':
                faction_color = values[1]
                #faction_controller = values[2] #'cpu', 'human' or 'all'
                resources = (int(values[3]), int(values[4]), int(values[5]))
                self.factions.append(Faction(faction_color, resources))
                log_message('Created faction: ' + faction_color)
            else:
                try:
                    x = int(values[1])
                    y = int(values[2])
                    owner = values[3]
                except:
                    error_message('Illegal line in game level config')
                    continue
                if values[0] == 'base':
                    self.objects.append(Base(x, y, owner, self))
                    log_message('Created base: ' + owner)
                    self.block_tile_rectangle((x, y, 3, 2))
                    self.unblock_tile((x, y + 1))
                    self.unblock_tile((x + 2, y + 1))
                elif values[0] == 'shrine':
                    shrine_type = values[4]
                    self.objects.append(Shrine(x, y, owner, shrine_type))
                    log_message('Created shrine: ' + owner)
                    self.block_tile_rectangle((x, y, 3, 2))
                    self.unblock_tile((x + 1, y + 1))
                elif values[0] == 'rock' or values[0] == 'paper' \
                or values[0] == 'scissors':
                    self.objects.append(Unit(x, y, owner, values[0], values[4]))
                    log_message('Created unit: ' + values[0] + ' for ' + owner)

        level_file.close()
        return

    def block_tile_rectangle(self, rect):
        """Make a rectangular area of tiles unpassable.
        Rect is a tuple of four values, NOT pygame rect"""
        x = rect[0]
        y = rect[1]
        target_x = rect[0] + rect[2]
        target_y = rect[1] + rect[3]
        while x < target_x:
            y = target_y - rect[3]
            while y < target_y:
                if (not self.tiles.has_key((x, y))) or \
                self.tiles[(x, y)] != TILE_WALL:
                    self.tiles[(x, y)] = TILE_BLOCKED
                y += 1
            x += 1
        return

    def unblock_tile(self, coords):
        """Make a tile passable."""
        del self.tiles[coords]
        return

    def render(self, screen, offset_x = 0, offset_y = 0):
    
        pygame.draw.rect(screen, COLOR_GUI_BG, pygame.Rect(offset_x, offset_y,\
        self.width * TILE_WIDTH, self.height * TILE_HEIGHT), 2 )

        x = 0
        while x < self.width:
            y = 0
            while y < self.height:
                if self.tiles.has_key((x, y)) and self.tiles[(x, y)] == TILE_WALL:
                    render_image(screen, 'tile_wall.png', \
                    (x * TILE_WIDTH + offset_x, y * TILE_HEIGHT + offset_y))
                y += 1
            x += 1

        for o in self.objects:
            o.render(screen, offset_x, offset_y)

        if self.selected_object != None:
            self.selected_object.render_infobox(screen, self)

        return

    def increase_resources(self, faction_color):
        """Server side: increase faction resources
        according to stuff on the map"""
        for f in self.factions:
            if (f.get_color() == faction_color):
                res_inc = self.get_resource_increase(faction_color)
                f.add_resources(res_inc)
        return

    def get_resource_increase(self, faction_color):
        res = MONEY_ZERO
        for o in self.objects:
            if o.owner == faction_color:
                new_res = o.produce_resources()
                res = (res[0] + new_res[0], res[1] + new_res[1], res[2] + new_res[2])

        return res

    def commandeer_faction(self):
        """Finds a faction with no commander and marks it
        as having a commander. Returns faction color.
        For server-side faction handling."""
        for f in self.factions:
            if not f.has_commander():
                return f.faction_color
        return NO_AVAILABLE_FACTIONS

    def get_faction(self, faction_color):
        """Find a faction object of given color.
        If no faction is found, returns None.
        For client and server-side faction handling."""
        for f in self.factions:
            if f.faction_color == faction_color:
                return f
        return None
        
    def deselect(self):
        if self.selected_object != None:
            self.selected_object.selected = False
        self.selected_object = None
        return

    def select(self, pos, faction = None):
        self.deselect()
        o = self.find(pos, faction)
        if o != None:
            o.selected = True
            self.selected_object = o
            return True
        return False

    def select_by_tile_coords(self, coords, faction = None):
        self.deselect()
        o = self.find_by_tile_coords(coords, faction)
        if o != None:
            o.selected = True
            self.selected_object = o
            return True
        return False

    def find(self, pos, faction = None):
        found = None
        for o in self.objects:
            if o.rect.collidepoint(pos) and (faction == None or o.owner == faction):
                if o.object_name == 'rock' or \
                o.object_name == 'paper' or o.object_name == 'scissors':
                    return o
                found = o
        return found

    def find_by_tile_coords(self, coords, faction = None):
        for o in self.objects:
            if o.x == coords[0] and o.y == coords[1] and (faction == None or o.owner == faction):
                return o
        return None

    def use_infobox_gui(self, event_pos, res):
        if self.selected_object != None and \
        self.selected_object.object_name == 'base':
            return self.selected_object.use_infobox_gui(event_pos, res)

        return None

    def toggle_twist(self, faction = None):
        """Toggle twist. faction is a string containing the faction color"""
        if faction == None:
            #Server side way of things
            self.twisted = True
        else:
            self.twisted = not self.twisted
            buyer = self.get_faction(faction)
            if self.twisted:
                if not buyer.use_resources(COST_TWIST):
                    self.twisted = not self.twisted
                    log_message('Not enough resources to twist: need 50 each')
                    return None
            else:
                buyer.add_resources(COST_TWIST)

        for o in self.objects:
            o.toggle_twist()
        return Action(ACTION_TWIST)

    def move(self, action_type):
        """Move selected unit as specified by action type (str)"""
        if self.selected_object != None:
            x = self.selected_object.x
            y = self.selected_object.y
            if self.selected_object.move(action_type):
                if self.tiles.has_key \
                ((self.selected_object.x, self.selected_object.y)) or \
                self.selected_object.x < 0 or self.selected_object.y < 0 \
                or self.selected_object.x == self.width \
                or self.selected_object.y == self.height:

                    #Collision with an unpassable tile or map edge detected
                    log_message('Collision')
                    self.selected_object.x = x
                    self.selected_object.y = y
                    self.selected_object.moves_used -= 1
                else:
                    for o in self.objects:
                        if o != self.selected_object \
                        and o.x == self.selected_object.x \
                        and o.y == self.selected_object.y:
                            self.selected_object.x = x
                            self.selected_object.y = y
                            self.selected_object.moves_used -= 1
                            if self.selected_object.owner != o.owner:
                                self.selected_object.attack(o)
                                o.attack(self.selected_object)
                    return Action(action_type, (x, y))
        return None

    def buy(self, action, faction):
        """Buy units as specified by action (an Action object).
        Faction is the faction color as a string."""
        occupier = self.find_by_tile_coords(action.coords)

        if action.action_type == ACTION_BUY_ROCK:
            buy_type = 'rock'
        elif action.action_type == ACTION_BUY_PAPER:
            buy_type = 'paper'
        else:
            buy_type = 'scissors'

        if occupier != None and (occupier.owner != faction or \
        occupier.unit_class != buy_type):
            return None

        cost = action.amount * Base.COST[buy_type][0]
        cost_res = MONEY_ZERO

        if Base.COST[buy_type][1] == 0:
           cost_res = (cost, 0, 0)
        elif Base.COST[buy_type][1] == 1:
           cost_res = (0, cost, 0)
        else:
           cost_res = (0, 0, cost)

        buyer = self.get_faction(faction)
        buyer.use_resources(cost_res)

        if occupier == None:
            unit = Unit(action.coords[0], action.coords[1], faction, buy_type, action.amount)
            unit.startamount = 0
            self.objects.append(unit)
        else:
            unit = occupier
            occupier.amount += action.amount
        unit.moves_used = unit.max_moves
        return action

    def new_turn(self):
        log_message('New turn')

        self.twisted = False
        
        winner = None

        for o in self.objects:
            o.new_turn()
            if o.object_name == 'base':
                #Conquering bases
                occupier_1 = self.find_by_tile_coords((o.x, o.y + 1))
                occupier_2 = self.find_by_tile_coords((o.x + 2, o.y + 1))
                if occupier_1 != None:
                    if occupier_2 == None or occupier_2.owner == occupier_1.owner:
                        o.owner = occupier_1.owner
                if occupier_2 != None:
                    if occupier_1 == None or occupier_2.owner == occupier_1.owner:
                        o.owner = occupier_2.owner
                if winner == None:
                    winner = o.owner
                else:
                    if o.owner != winner:
                        winner = ''
            elif o.object_name.startswith('shrine'):
                #Conquering shrines
                occupier_1 = self.find_by_tile_coords((o.x + 1, o.y + 1))
                if occupier_1 != None:
                    o.owner = occupier_1.owner

        for o in self.objects:
            if o.dead:
                #Removing dead objects
                log_message('removed: ' + o.object_name)
                self.objects.remove(o)

        if winner == None:
            winner = ''

        return winner

    def reset_last_turn(self):
        log_message('Resetting turn')
        for o in self.objects:
            o.reset_last_turn()
        return