"""Main game server module. Used for both single player and multiplayer."""

import socket
import select
import time
from threading import Thread

from locals import *

import data

from variables import Variables
from log import log_message

from unit import Unit
from faction import Faction
from level import Level
from turn import Turn
from game_round import Game_Round
from action import Action

class Server(Thread):

    def __init__(self, is_public, port = PORT_DEFAULT, key = '', level_name = '2p_clash'):
        Thread.__init__(self)
        self.is_public = is_public
        self.port = port
        self.level = Level(level_name)
        self.factions = self.level.factions
        self.key = key
        return

    def add_faction(self, controller = PLAYER, controller_address = None):
        return

    def run(self):
        try:
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            host = ''
            server.bind((host, self.port))
            server.listen(SERVER_BACKLOG)
            s_input = [server]
        except socket.error, (value,message):
            if sock:
                sock.close()
            error_message("Server: Could not open socket: " + message)
            return

        log_message("Server: Running")

        terminated = False
        terminating = False
        tried_to_terminate = 0

        self.event_queue = []

        self.current_clients = []

        running_id = 0

        self.current_round = None
        self.round_data_received = False

        while not terminated:

            #To give processing time to other processes and threads:
            time.sleep(1.0/SERVER_RATE)

            if terminating:
                #Trying to terminate server:
                log_message('Server terminating: still has clients: ' \
                + str(len(self.current_clients)))
                tried_to_terminate += 1
                #Waits for clients to disconnect in an orderly fashion for
                #2 seconds, then goes down automatically.
                log_message('timeout not yet reached: ' + str(tried_to_terminate) + ' < ' + str(SERVER_RATE*2) )
                if tried_to_terminate > SERVER_RATE*2:
                    terminated = True

            if self.current_round == None:
                #Round has ended or game is beginning
                self.current_round = Game_Round()
            else:
                if self.round_data_received:
                    event = self.current_round.process_round(self.level)
                    targets = []
                    if self.current_round.finished:
                        self.current_round = None
                        self.round_data_received = False
                        log_message('Server: Round ending')
                        for f in self.factions:
                            log_message('Server: adding resources for faction ' + f.get_color())
                            self.level.increase_resources(f.get_color())
                        for c in self.current_clients:
                            targets = [c]
                            for f in self.factions:
                              if (f.get_color() == c[3]):
                                res = f.get_resources()
                                str_res = str(res[0]) + ' ' + \
                                str(res[1]) + ' ' + str(res[2])
                                self.event_queue.append((GAME_COMMAND_SET_RES \
                                + ' ' + str_res, targets))
                    elif event != '':
                        for c in self.current_clients:
                            targets.append(c)
                        self.event_queue.append((event, targets))

            #Begin network input processing

            #Here's where the game gets stuck:
            inputready,outputready,exceptready = select.select(s_input,[],[])

            for s in inputready:

                if (s == server):
                    client, address = server.accept()
                    log_message("Server: " + str(address[0]) + \
                    " trying to connect")
                    if self.is_public or address[0] == '127.0.0.1':
                        s_input.append(client)
                        self.current_clients.append((address[0], -1, \
                        'Anonymous', 'observer'))
                        log_message("Server accepting messages from " + \
                        address[0])
                else:
                    data = s.recv(PACKET_MAX_SIZE)
                    if not data:
                        s.close()
                        s_input.remove(s)
                    else:
                        #log_message("Server received: " + data)
                        #Begin server side connection management and basic
                        #traffic processing
                        if (data.startswith(COMMAND_TERMINATE) \
                        and data.endswith(self.key)):
                            #Adminstrating client closes off the server
                            s.send(RESPONSE_TERMINATE)
                            targets = []
                            for c in self.current_clients:
                                targets.append(c)
                            self.event_queue.append((COMMAND_KILLYOURSELF, \
                             targets))
                            terminating = True
                        elif (data.startswith(COMMAND_PING)):
                            #Ping-pong
                            s.send(RESPONSE_PING)
                        elif (data.startswith(COMMAND_GET_ID)):
                            #A client requests to register a client ID
                            #(to separate clients on the same host)
                            try:
                                client_nick = data[(len(COMMAND_GET_ID) + 1):]
                            except:
                                client_nick = 'Anonymous'
                            id_sent = False
                            for c in self.current_clients:
                                if (c[1] == -1):
                                    #This should also check the source IP,
                                    #I'll have to find out how to do that
                                    #- or rather create a client-specific
                                    #authentication key
                                    running_id += 1
                                    s.send(RESPONSE_GET_ID + ' ' + \
                                    str(running_id))
                                    client_nick = client_nick[0:NICK_MAX_LEN]
                                    client_nick = client_nick.strip()
                                    self.current_clients.append((c[0], \
                                    running_id, client_nick, 'observer'))
                                    self.current_clients.remove(c)
                                    self.serve_chat_msg(client_nick + \
                                    " joined the fun (/nick to change nick)")
                                    id_sent = True
                            if not id_sent:
                                s.send(RESPONSE_GET_ID + ' ' + '0')
                        elif (data.startswith(COMMAND_DISCONNECT)):
                            #A client requests to disconnect
                            s.send(RESPONSE_DISCONNECT)
                            values = data.split()
                            for c in self.current_clients:
                                if c[1] == int(values[1]):
                                    self.serve_chat_msg(c[2] + \
                                    " disconnected")

                                    log_message('Server: disconnecting ID ' \
                                     + str(values[1]) + ' succesful.')

                                    self.current_clients.remove(c)
                            if terminating and (len(self.current_clients) == 0):
                                terminated = True

                        else:
                            #Begin server side game and chat related traffic
                            #processing
                            c_client_id, empty, command = data.partition(' ')
                            command, empty, params = command.partition(' ')
                            c_client_id = int(c_client_id)
                            nick = self.authenticate_client(s, c_client_id)
                            if not nick:
                                s.send(RESPONSE_AUTH_FAILURE + data)
                                continue
                            if (command == COMMAND_SET_NICK):
                                #A client changes the nick
                                new_nick = (params[0:NICK_MAX_LEN]).strip()
                                for c in self.current_clients:
                                    if c[1] == c_client_id:
                                        self.current_clients.append((\
                                        c[0], c[1], new_nick, c[3]))
                                        self.current_clients.remove(c)
                                if nick != new_nick:
                                    self.serve_chat_msg(nick + \
                                    " changed nick to " + new_nick)
                                s.send(command + GENERIC_RESPONSE)
                            elif (command == COMMAND_CHAT):
                                #A client sends a chat message to the server
                                targets = []
                                for c in self.current_clients:
                                    targets.append(c)
                                log_message("Server: broadcasting chat msg")
                                self.serve_chat_msg("<" + nick + "> " + params)
                                s.send(command + GENERIC_RESPONSE)
                            elif (command == COMMAND_QUERY):
                                #A client queries the server for new data coming
                                #from the server to the client
                                events_for_client = 0
                                event_to_send = None
                                for event in self.event_queue:
                                    targets = event[1]
                                    for target in targets:
                                        if target[1] == c_client_id:
                                             events_for_client += 1
                                             if event_to_send == None:
                                                 event_to_send = str(event[0])
                                                 targets.remove(target)

                                for event in self.event_queue:
                                    targets = event[1]
                                    if len(targets) == 0:
                                        self.event_queue.remove(event)

                                if (events_for_client > 0):
                                    s.send(command + GENERIC_RESPONSE + " " + \
                                    str(events_for_client - 1) + " " + \
                                    event_to_send)
                                else:
                                    s.send(command + GENERIC_RESPONSE + " " + \
                                    str(events_for_client) + " " + "none")
                            else:
                                #Process actual game-related data and respond
                                #to unknown commands
                                game_data = self.process_game_command(command,\
                                params, c_client_id)
                                if game_data:
                                    s.send(command + GENERIC_RESPONSE + ' ' + \
                                    game_data)
                                else:
                                    #echo response on unknown command:
                                    s.send(RESPONSE_UNKNOWN_COMMAND + data)
            #end network input processing
            pass

        #Server main loop has terminated:
        client.close()
        server.close()
        log_message("Server: Terminated")

        return

    def authenticate_client(self, sock, client_id):
        for c in self.current_clients:
            if (c[1] == client_id):
                return c[2]
        return False

    def process_game_command(self, command, params, client_id = -1):
        """Process game related data sent in by the client"""
        if (command == GAME_COMMAND_ASK_LEVEL):
            return (self.level.name + ' ' + self.level.file_hash)
        elif (command == GAME_COMMAND_ASK_FACTION):
            faction_color = self.level.commandeer_faction()
            for c in self.current_clients:
                if (c[1] == client_id):
                    self.current_clients.append((c[0], c[1], c[2], faction_color))
                    self.current_clients.remove(c)
            return faction_color
        elif (command == GAME_COMMAND_SEND_TURN):
            if self.current_round != None:
                faction_color = 'observer'
                for c in self.current_clients:
                    if (c[1] == client_id):
                        faction_color = c[3]

                turn_ready = self.current_round.add_action_from_msg(params, \
                self.level.get_faction(faction_color))

                if turn_ready:
                    self.serve_chat_msg('Faction ' + faction_color + \
                    ' has finished its turn.')

                if self.current_round.amount_factions_received() \
                == len(self.current_clients):
                    self.round_data_received = True

            return ' '
        return False

    def serve_chat_msg(self, message, targets = None):
        if targets == None:
            targets = []
            for c in self.current_clients:
                targets.append(c)
        self.event_queue.append((COMMAND_CHAT + " " + \
        time.strftime('%H:%M') + " " + message, targets))
        return True


