""" This module contains generic utility functions.
The comments give a basic idea of each function's purpose,
and they can also be found from the WWW documentation."""

import pygame
from pygame.locals import *
from string import lower
import os
import codecs
import datetime

from locals import *

import data

from variables import Variables

from log import error_message, log_message

class Util:
    """A storage class for cached images, variables used by the
    utility functions and such."""
    pygame.font.init()
    cached_fonts = {}
    cached_text_images = {}
    cached_images = {}
    fade_state = FADE_STATE_BLACK
    edit_cursor = 0
    blackscreen = pygame.Surface((SCREEN_WIDTH, SCREEN_HEIGHT))

def get_config_path():
    """Return a path for saving config in the user's home directory.
    It's compatible with both UNIX-likes (environment variable HOME)
    and Windows (environment variable APPDATA).
    Prints error messages if unsuccesful, and reverts to a "saves" directory
    under the data directory instead.
    BE SURE TO CHANGE THE GAME_NAME_SHORT CONSTANT BY EDITING locals.py
    BEFORE USING THIS FUNCTION!"""
    path_name = ""
    try:
        path_name = os.path.join(os.environ["HOME"], "." + GAME_NAME_SHORT)
    except:
        error_message("Couldn't find environment variable HOME, " +
        "reverting to APPDATA. This is normal under Windows.")
        try:
            path_name = os.path.join(os.environ["APPDATA"], GAME_NAME_SHORT)
        except:
            error_message("Couldn't get environment variable for home " +
            "directory, using data directory instead.")
            path_name = data.filepath("saves")
    if not os.path.exists(path_name):
        os.mkdir(path_name)
    return path_name

def parse_config():
    """Parse a config file stored in the location given by get_config_path().
    The "sound" and "fullscreen" values are parsed by default, but you can
    add your own by editing the code. The parsed values are stored in the 
    Variables class of the variables module."""
    #Default values for configuration (example: sound is on):
    Variables.vdict["sound"] = True
    Variables.vdict["fullscreen"] = False
    Variables.vdict["player_name"] = "Anonymous"

    #Setting this to True will enable printing error messages to the console.
    #Might be useful during debugging.
    #You may set it to False in a release version to disable console output.
    #Messages will still be added to the message log file.
    Variables.vdict["verbose"] = True

    file_path = os.path.join(get_config_path(), "config.txt")
    try:
        conffile = codecs.open(file_path, "r", "utf_8")
        for line in conffile.xreadlines():
            if is_config_line(line):
                values = line.split("\t")

                #Parsing configuration values:

                if values[0] == "sound" or values[0] == "fullscreen":
                    Variables.vdict[values[0]] = str_to_bool(values[1])
                    #The "sound" and "fullscreen" values are of type bool
                if values[0] == "player_name":
                    Variables.vdict[values[0]] = values[1]
                    #The player name value is of type str

                #Add your own value types here.

    except:
        if write_config():
            log_message("Created configuration file to " + file_path)

    return

def write_config():
    """Write the config stored in the Variables class to the configuration
    file. Prints an error message if unsuccessful."""
    file_path = os.path.join(get_config_path(), "config.txt")
    try:
        conffile = codecs.open(file_path, "w", "utf_8")

        #Variables to be saved:
        print >> conffile, "sound\t%s" % bool_to_str(Variables.vdict["sound"])
        print >> conffile, "fullscreen\t%s" % \
        bool_to_str(Variables.vdict["fullscreen"])
        print >> conffile, "player_name\t%s" % Variables.vdict["player_name"]

    except:
        error_message("Could not write configuration file to " + file_path)
        return False

    return True

def write_log():
    """Write the message log to disk."""
    file_path = os.path.join(get_config_path(), "log.txt")
    old_log = ""
    if os.path.exists(file_path):
        conffile = codecs.open(file_path, "r", "utf_8")
        count = 0
        for line in conffile.xreadlines():
            old_log = old_log + line
            count += 1
            if count > MAX_OLD_LOG_LINES:
                break
    if Variables.vdict.has_key("log"):
        try:
            conffile = codecs.open(file_path, "w", "utf_8")
            print >> conffile, "Log updated " + str(datetime.date.today())
            print >> conffile, Variables.vdict["log"]
            print >> conffile, ""
            print >> conffile, old_log
        except:
            error_message("Could not write log file to " + file_path)
            return False
    return True

def is_config_line(line):
    """Return true if the line is not empty or a comment starting with #"""
    if line.strip() == "":
        return False
    if (line.strip()).startswith("#"):
        return False
    return True

def str_to_bool(string):
    """Convert string to bool"""
    formatted = string.strip()
    formatted = lower(formatted)
    return (formatted == "true" or formatted == "1" or formatted == "on")

def bool_to_str(bool):
    """Convert bool to string"""
    if bool:
        return "on"
    else:
        return "off"

def render_image(screen, file_name, coords = (0,0)):
    """Draw images on screen. Handles caching of images loaded from
    the file system. Images must be located in the pictures subdirectory
    of the data directory"""
    image = None
    if not (Util.cached_images.has_key(file_name)):
        try:
            i_path = data.filepath(os.path.join("pictures", file_name))
            image = Util.cached_images[file_name] \
            = pygame.image.load(i_path).convert()
        except:
            error_message("Image file not found: " + file_name)
            return
    else:
        image = Util.cached_images[file_name]
    screen.blit(image, coords)
    rect = image.get_rect()
    rect.topleft = coords
    return rect

def render_text(string, size = FONT_SIZE, color = None, bgcolor = None):
    """Render fancy-looking text. The function also handles caching of text
    images. Returns a pygame surface containing the rendered text.
    The text is rendered with slight edges to make it look more readable on a
    colorful background.

    The constant colors can be found from locals.py."""

    #Backwards compatibility with Splush versions <= 0.9.5:
    if type(size) != type(FONT_SIZE):
        if color != None:
            bgcolor = (color[0], color[1], color[2])
        if size != None:
            color = (size[0], size[1], size[2])
        size = FONT_SIZE
    if color == None:
        color = COLOR_GUI
    if bgcolor == None:
        bgcolor = COLOR_GUI_BG
    #End backwards compatibility


    if Util.cached_text_images.has_key(string + str(color) + \
    str(bgcolor) + ' ' + str(size)):
        final_image = Util.cached_text_images[string + str(color) +
        str(bgcolor) + ' ' + str(size)]
    else:
        if Util.cached_fonts.has_key(size):
            font = Util.cached_fonts[size]
        else:
            font = Util.cached_fonts[size] = pygame.font.Font(data.filepath(\
            os.path.join("misc", "Vera.ttf")), size)

        text_image_bg = font.render(string, True, bgcolor)
        text_image_fg = font.render(string, True, color)
        rect = text_image_bg.get_rect()
        final_image = pygame.Surface((rect.width + 2, rect.height + 2))\
        .convert_alpha()


        final_image.fill((0,0,0,0))

        final_image.blit(text_image_bg, rect)
        final_image.blit(text_image_bg, (1,0))
        final_image.blit(text_image_bg, (2,0))

        final_image.blit(text_image_bg, (0,2))
        final_image.blit(text_image_bg, (1,2))
        final_image.blit(text_image_bg, (2,2))

        final_image.blit(text_image_fg, (1,1))
        Util.cached_text_images[string + str(color) + str(bgcolor) + \
        ' ' + str(size)] = final_image
    return final_image

def render_text_dialogue(screen, string, phase):
    """Render text partially.
    For fancy dialogue display.
    The phase value is the amount of characters shown.
    -1 phase means that the whole string is visible.
    """
    if phase == -1:
        phase = len(string)

    rendered_string = string[0:phase]
    string_image = render_text(rendered_string)
    rect = string_image.get_rect()
    rect.centerx = SCREEN_WIDTH / 2
    rect.centery = SCREEN_HEIGHT / 2
    screen.blit(string_image, rect)

    if phase < len(string):
        phase += 1
        #You can add a sound effect here, if you want, example code:
        #play_sound("click")
    else:
        return -1

    return phase

def clear_text_cache():
    """Clears image cache generated by the render_text function.
    The cache can take a lot of space in memory, so it's recommended to use
    this function to empty it occasionally. Especially if you use large fonts
    or use render_text to render text input from the user.
    Font cache isn't cleared."""
    Util.cached_text_images = {}
    return

def set_fade_state(fade_state):
    """Set the fade state used by fade_to_black"""
    Util.fade_state = fade_state
    return

def fade_to_black(screen, fade_target):
    """Fades the screen to black. Both fade-in and fade-out.
    Returns true if the fading has finished.
    fade_target should be an integer (0-255)
    * 255 = FADE_STATE_BLACK : The display is all black
    * 0 = FADE_STATE_NONE : The display is not faded at all"""
    if Util.fade_state > fade_target:
        Util.fade_state += int(255 / (FPS * FADE_IN))
        if Util.fade_state < fade_target:
            Util.fade_state = fade_target
    if Util.fade_state < fade_target:
        Util.fade_state -= -int(255 / (FPS * FADE_OUT))
        if Util.fade_state > fade_target:
            Util.fade_state = fade_target
    if Util.fade_state > FADE_STATE_NONE:
        Util.blackscreen.set_alpha(Util.fade_state)
        screen.blit(Util.blackscreen, screen.get_rect())
    return (Util.fade_state == fade_target)

def apply_fullscreen_setting(screen):
    """Apply the fullscreen setting stored in variables to the screen.
    screen is a pygame standard surface.
    Contents of the screen should not be affected."""
    mode = 0
    if Variables.vdict["fullscreen"]:
        mode ^= FULLSCREEN
    tmp = pygame.Surface((SCREEN_WIDTH,SCREEN_HEIGHT)).convert()
    tmp.blit(screen,(0,0))
    screen = pygame.display.set_mode((SCREEN_WIDTH,SCREEN_HEIGHT),mode)
    screen.blit(tmp,(0,0))
    pygame.display.flip()
    return

def ask_input(screen, label = 'Input', msg = ''):
    box_left = SCREEN_WIDTH / 2 - 150
    box_top = SCREEN_HEIGHT / 2 - 75
    finished = False
    clock = pygame.time.Clock()
    cursor = Util.edit_cursor = len(msg)
    while not finished:
        clock.tick(FPS)
        for event in pygame.event.get():
            if event.type == KEYDOWN and event.key == K_RETURN:
                finished = True
                return msg
            if event.type == KEYDOWN and event.key == K_ESCAPE:
                finished = True
                return None
            else:
                msg, cursor = str_editor(msg, event)
        render_image(screen, 'menu-background.png', (box_left, box_top))
        label_img = render_text(label)
        screen.blit(label_img, (box_left + 5, box_top + 5))

        msg_show = add_cursor(msg, cursor)
        msg_img = render_text(msg_show)
        screen.blit(msg_img, (box_left + 5, box_top + 35))
        pygame.display.flip()
    clear_text_cache()
    return ''
    
def add_cursor(string, cursor):
    return string[0:cursor] + '|' + string[cursor:len(string)]

def str_editor(string, event):
    """Manipulate a string via keyboard events. Useful for text input.
    Only use this for one string at a time!
    Returns the edited string and the position of the cursor."""
    if (string == ''):
        Util.edit_cursor = 0
    if event.type == KEYDOWN:
        if event.key == K_HOME:
            Util.edit_cursor = 0
        elif event.key == K_END:
            Util.edit_cursor = len(string)
        elif event.key == K_LEFT:
            if Util.edit_cursor > 0:
                Util.edit_cursor -= 1
        elif event.key == K_RIGHT:
            if Util.edit_cursor < (len(string)):
                Util.edit_cursor += 1
        elif event.key == K_BACKSPACE:
            if len(string) > 0:
                string = string[0:(Util.edit_cursor - 1)] + \
                string[Util.edit_cursor:len(string)]
                Util.edit_cursor -= 1
        elif event.key == K_DELETE:
            pass
        elif event.key == K_TAB or event.key == K_ESCAPE:
            pass
        elif Util.edit_cursor == len(string):
            string = string + event.unicode.encode('latin-1')
            Util.edit_cursor += 1
        else:
            string = string[0:Util.edit_cursor] + \
            event.unicode.encode('latin-1') + string[Util.edit_cursor:len(string)]
            Util.edit_cursor += 1

    return (string, Util.edit_cursor)